seed_centre = c(13.143085, -9.668144,12.012177)

library(readr)
library(purrr)
library(rgl)
library(reshape2)
library(scatterplot3d)
library(bio3d)

print("RDA matrix structure started")
plot_to_check = 0

data_compiled_assemblages <- read_csv("C:/data_compiled_assemblages.csv") # This is provided in the Seanoe database
#We remove undetermined data and polygons exhibiting bare substratum for three time steps.
data_compiled_assemblages=data_compiled_assemblages[-which(data_compiled_assemblages$Assemblage_2015=="i"|data_compiled_assemblages$Assemblage_2018=="i"|data_compiled_assemblages$Assemblage_2020=="i"),]
data_compiled_assemblages=data_compiled_assemblages[-which(data_compiled_assemblages$Megafauna_substratum_2015=="su1"&data_compiled_assemblages$Megafauna_substratum_2018=="su1"&data_compiled_assemblages$Megafauna_substratum_2020=="su1"),]

ID_tot_original = data_compiled_assemblages$ID_tot
data_compiled_assemblages$ID_tot=rownames(data_compiled_assemblages)

data_assemblages=as.data.frame(data_compiled_assemblages)
level_assemblage=levels(as.factor(data_assemblages[,5]))

Matrix_RDA = as.data.frame(matrix(nrow = nrow(data_assemblages), ncol = 18+3+2+1))
colnames(Matrix_RDA) = c("Side", "Centroid_x", "Centroid_y", "Centroid_z", "Large B. azoricus", "Medium B. azoricus", "Small B. azoricus", "white hydrothermal deposit", "Empty shell", "Hydrozoan sp.", "Amorphous sulphur", "Bare substratum", "Zoanthid sp.", "Microbial mat", "S. mesatlantica", "Phymorhynchus sp.","M. fortunata", "P. smaragdina", "Main smoker", "Secondary smoker", "Diffusion zone", "Flange", "Hydrothermal topographic change", "Non hydrothermal topographic change")
Matrix_RDA$`S. mesatlantica`=0
Matrix_RDA$`Phymorhynchus sp.`=0

Matrix_RDA_2015=Matrix_RDA[,5:ncol(Matrix_RDA)]
Matrix_RDA_2018=Matrix_RDA[,5:ncol(Matrix_RDA)]
Matrix_RDA_2020=Matrix_RDA[,5:ncol(Matrix_RDA)]

#Extract polygon coordinates
x = map_df(strsplit(data_compiled_assemblages$x, ","), ~as.data.frame(t(.)))
y = map_df(strsplit(data_compiled_assemblages$y, ","), ~as.data.frame(t(.)))
z = map_df(strsplit(data_compiled_assemblages$z, ","), ~as.data.frame(t(.)))

Matrix_RDA[,1]=data_assemblages[,1]
for (i in 1:nrow(data_assemblages)){
  Matrix_RDA$x[i] = mean(as.numeric(x[i,]), na.rm=TRUE)
  Matrix_RDA$y[i] = mean(as.numeric(y[i,]), na.rm=TRUE)
  Matrix_RDA$z[i] = mean(as.numeric(z[i,]), na.rm=TRUE)
  # print(as.character(i))
}

# Get matrix of polygon points
vector_x = as.numeric(unlist(t(x)))
polygon_indices = x
for (i in 1:nrow(x)){polygon_indices[i,]=rep(i,ncol(x))}
# Ok pas bon c'est pour cela qu'on a une erreur...
polygon_indices = as.data.frame(as.vector(t(polygon_indices)))
#polygon_indices_2 = polygon_indices[!is.na(vector_x),1]
#polygon_indices = as.numeric(polygon_indices_2)
vector_y = as.numeric(unlist(t(y)))
#vector_y = vector_y[!is.na(vector_y)]
vector_z = as.numeric(unlist(t(z)))
#vector_z = vector_z[!is.na(vector_z)]
polygon_indices = cbind(polygon_indices, vector_x, vector_y, vector_z)
polygon_indices = polygon_indices[!is.na(polygon_indices$vector_x),]
colnames(polygon_indices)=c("polygon_indices","x","y","z")

# Get the surface of max aggregation
ordered_data = data_assemblages[with(data_assemblages,order(-Surface)),]
ordered_data = ordered_data[1:10,]

# Get nearest polygons of the 3D model centre
dist = dist.xyz(seed_centre, polygon_indices[,2:4])
polygon_indices = as.data.frame(polygon_indices)
polygon_indices = cbind(polygon_indices, t(dist))
polygon_indices = polygon_indices[order(polygon_indices$`t(dist)`, decreasing = TRUE),]
# Centroid of polygons
#start_index = polygon_indices[which.min(dist),1]
#polygon_of_start = data_assemblages[polygon_indices[which.min(dist),1],]
polygon_indices_backup = polygon_indices


# Loop to aggregate
max_polygon_surface = mean(ordered_data$Surface)
data_groups = as.data.frame(matrix(nrow = nrow(data_assemblages), ncol = 1))
polygon_indices = polygon_indices_backup
centroids = cbind(Matrix_RDA[,25:27],as.data.frame(seq(1:nrow(Matrix_RDA))))
colnames(centroids) = c("x", "y", "z", "polygon_index")
log=as.vector(0)
# We take the farthest polygon
for (i in 1:nrow(data_assemblages)){
  # We select the polygon of reference (i.e. the farthest point from the 3D model centre), to which other polygons will be associated.
  # polygon_indices = polygon_indices_backup_i
  
  index_polygon_ref = polygon_indices$polygon_indices[1]
  vertices_ref = polygon_indices[which(polygon_indices$polygon_indices==index_polygon_ref),] # We select the farthest from the centre of the 3D model
  polygon_ref = data_assemblages[index_polygon_ref,]
  centroid_ref = centroids[which(centroids$polygon_index == index_polygon_ref),]
  centroids = centroids[-which(centroids[,4]==index_polygon_ref),] # We remove the centroid of the polygon of reference
  polygon_indices = polygon_indices[-which(polygon_indices$polygon_indices == index_polygon_ref),]
  data_groups[index_polygon_ref,1] = i
  log=c(log,index_polygon_ref)
  
  surface_total =  polygon_ref$Surface
  if (surface_total  < max_polygon_surface){ #if the polygon is not bigger than the treshold
    # LOOP FROM HERE
    for(n in 1:nrow(data_assemblages)){
      # We determine which polygon will be the candidate based on the centroid distance (i.e. the closest so minimum).
      dist_centroids = cbind(t(dist.xyz(centroid_ref[1:3], centroids[,1:3])),centroids[,4])
      index_polygon_cand = centroids[which(centroids$polygon_index == dist_centroids[which.min(dist_centroids[,1]),2]),4]
      polygon_candidate = data_assemblages[index_polygon_cand,]
      surface_total = polygon_candidate$Surface + surface_total
      dist_vertices = dist.xyz(vertices_ref[,2:4],polygon_indices[which(polygon_indices$polygon_indices == index_polygon_cand),2:4])
      if(surface_total < max_polygon_surface && min(dist_vertices)<sqrt(max_polygon_surface)){ 
        data_groups[index_polygon_cand,1] = i
        centroids = centroids[-which(centroids[,4]==index_polygon_cand),]
        polygon_indices = polygon_indices[-which(polygon_indices$polygon_indices == index_polygon_cand),]
      }else if(surface_total > max_polygon_surface){
        break}
    }
  }
}
if(sum(is.na(data_groups[,1]))==0){print("Ok, all polygons have been assigned to a group.")}

data_groups_backup = data_groups 

aggregated_cover_stats = aggregate(data_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
quantile(aggregated_cover_stats$x, c(.01, .05, .1, .25, .5, .75, .95))
max(aggregated_cover_stats$x)
hist(aggregate(data_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)[,2])
aggregated_cover_stats_backup=aggregated_cover_stats

centroids = cbind(Matrix_RDA[,25:27],as.data.frame(seq(1:nrow(Matrix_RDA))))
colnames(centroids) = c("x", "y", "z", "polygon_index")
data_groups=data_groups_backup
aggregated_cover_stats=aggregated_cover_stats_backup

######### Refinement 1

subset_aggregated_cover = aggregated_cover_stats[aggregated_cover_stats$x< 1.2,] # criterion for max size of polygons that will be considered for tiling
centroids_without_polygons_to_reallocate = as.data.frame(centroids[-which(data_groups$V1 %in% as.vector(subset_aggregated_cover$Category)),])
criterion_max_group_surface = 2.75  # criterion for max tile surface

for(j in 1:nrow(subset_aggregated_cover)){ # On loop sur les groupes ? r?assigner
  centroids_without_polygons_to_reallocate = as.data.frame(centroids[-which(data_groups$V1 %in% as.vector(subset_aggregated_cover$Category)),])
  print(j)
  index_group_ref = subset_aggregated_cover$Category[j]
  polygons_ref = data_assemblages[which(data_groups$V1==index_group_ref),]
  if(sum(polygons_ref$Surface) != aggregated_cover_stats$x[aggregated_cover_stats$Category==index_group_ref]){print("Something went wrong, the surfaces do not match")
  }else{
    centroids_with_polygons_to_reallocate = centroids[which(data_groups$V1==index_group_ref),]
    dist_centroids = dist.xyz(centroids_with_polygons_to_reallocate[,1:3], centroids_without_polygons_to_reallocate[,1:3])
    
    for (i in 1:nrow(dist_centroids)){
      polygon_index_candidate = which.min(dist_centroids[i,])
      data_group_cand = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate]]
      dist_centroids[i,polygon_index_candidate] = NA
      polygon_index_candidate_2 = which.min(dist_centroids[i,])
      data_group_cand_2 = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate_2]]
      dist_centroids[i,polygon_index_candidate] = NA
      polygon_index_candidate_3 = which.min(dist_centroids[i,])
      data_group_cand_3 = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate_3]]
      
      
      if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){ # Criterion for assemblage maximal surface (max +0.5m?)
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
        
      }else if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand_2)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){
        print("better")
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand_2
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
      }else if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand_3)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){
        print("better")
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand_3
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
        
      }else{print("not assigned")
        print(as.character(aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand)]+polygons_ref$Surface[i]))}
    }
  }}
hist(aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)[,2])
quantile(aggregated_cover_stats$x, c(.01, .05, .1, .25, .5, .75, .95))


######### Refinement 2
subset_aggregated_cover = aggregated_cover_stats[aggregated_cover_stats$x< 1.3,] # criterion for max size of polygons that will be considered for tiling
centroids_without_polygons_to_reallocate = as.data.frame(centroids[-which(data_groups$V1 %in% as.vector(subset_aggregated_cover$Category)),])
criterion_max_group_surface = 2.75 # criterion for max tile surface

for(j in 1:nrow(subset_aggregated_cover)){ # On loop sur les groupes ? r?assigner
  centroids_without_polygons_to_reallocate = as.data.frame(centroids[-which(data_groups$V1 %in% as.vector(subset_aggregated_cover$Category)),])
  print(j)
  index_group_ref = subset_aggregated_cover$Category[j]
  polygons_ref = data_assemblages[which(data_groups$V1==index_group_ref),]
  if(sum(polygons_ref$Surface) != aggregated_cover_stats$x[aggregated_cover_stats$Category==index_group_ref]){print("Something went wrong, the surfaces do not match")
  }else{
    centroids_with_polygons_to_reallocate = centroids[which(data_groups$V1==index_group_ref),]
    dist_centroids = dist.xyz(centroids_with_polygons_to_reallocate[,1:3], centroids_without_polygons_to_reallocate[,1:3])
    
    for (i in 1:nrow(dist_centroids)){
      polygon_index_candidate = which.min(dist_centroids[i,])
      data_group_cand = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate]]
      dist_centroids[i,polygon_index_candidate] = NA
      polygon_index_candidate_2 = which.min(dist_centroids[i,])
      data_group_cand_2 = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate_2]]
      dist_centroids[i,polygon_index_candidate] = NA
      polygon_index_candidate_3 = which.min(dist_centroids[i,])
      data_group_cand_3 = data_groups$V1[centroids_without_polygons_to_reallocate$polygon_index[polygon_index_candidate_3]]
      
      
      if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){ # Criterion for assemblage maximal surface (max +0.5m?)
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
        
      }else if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand_2)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){
        print("better")
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand_2
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
      }else if((aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand_3)]+polygons_ref$Surface[i])<(criterion_max_group_surface)){
        print("better")
        data_groups$V1[as.numeric(polygons_ref$ID_tot[i])]=data_group_cand_3
        aggregated_cover_stats = aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)
        aggregated_cover_stats = aggregated_cover_stats[order(aggregated_cover_stats$x),]
        
      }else{print("not assigned")
        print(as.character(aggregated_cover_stats$x[which(aggregated_cover_stats$Category == data_group_cand)]+polygons_ref$Surface[i]))}
    }
  }}
hist(aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)[,2])
quantile(aggregated_cover_stats$x, c(.01, .05, .1, .25, .5, .75, .95))



# Prepare data to plot
# We sort it out according to data_groups
data_compiled_assemblages = cbind(data_compiled_assemblages, data_groups)
data_compiled_assemblages$ID_tot = ID_tot_original
#Extract polygon coordinates
x = map_df(strsplit(data_compiled_assemblages$x, ","), ~as.data.frame(t(.)))
y = map_df(strsplit(data_compiled_assemblages$y, ","), ~as.data.frame(t(.)))
z = map_df(strsplit(data_compiled_assemblages$z, ","), ~as.data.frame(t(.)))

#assign random colour code to plot
vector_color_groups = as.data.frame(cbind(data_compiled_assemblages$V1, data_compiled_assemblages$V1, data_compiled_assemblages$V1))
for (j in 1:3){
  for (i in levels(as.factor(data_compiled_assemblages$V1))){
    indices_groups = which(data_compiled_assemblages$V1==i)
    value = runif(1,0,255)
    for (n in 1:length(indices_groups)){
      vector_color_groups[indices_groups[n],j]=value}
  }
}

open3d()
par3d(windowRect = c(50, 50, 1500, 800))
#segments3d(x = as.numeric(unlist(strsplit(data$x[1], ","))), y = as.numeric(unlist(strsplit(data$y[1], ","))), z = as.numeric(unlist(strsplit(data$z[1], ","))))
for (i in 1:nrow(x)){
  x_tempo = as.numeric(unlist(x[i,]))
  x_tempo=x_tempo[!is.na(x_tempo)]
  x_tempo_2 = x_tempo
  #x_tempo_2[c(1,length(x_tempo_2))]=NA
  #x_tempo = as.vector(rbind(x_tempo_2, x_tempo))
  x_tempo = x_tempo[!is.na(x_tempo)]
  y_tempo = as.numeric(unlist(y[i,]))
  y_tempo=y_tempo[!is.na(y_tempo)]
  y_tempo_2 = y_tempo
  #y_tempo_2[c(1,length(y_tempo_2))]=NA
  #y_tempo = as.vector(rbind(y_tempo_2, y_tempo))
  y_tempo = y_tempo[!is.na(y_tempo)]
  z_tempo = as.numeric(unlist(z[i,]))
  z_tempo=z_tempo[!is.na(z_tempo)]
  z_tempo_2 = z_tempo
  #z_tempo_2[c(1,length(z_tempo_2))]=NA
  #z_tempo = as.vector(rbind(z_tempo_2, z_tempo))
  z_tempo = z_tempo[!is.na(z_tempo)]
  #polygon3d(x = x_tempo, y = y_tempo, z = z_tempo, fill=F)
  #polygon3d(x = x_tempo, y = y_tempo, z = z_tempo, col=rgb(0+255*data_groups$V1[i]/max(data_groups$V1), 255-255*data_groups$V1[i]/max(data_groups$V1), 255-255*data_groups$V1[i]/max(data_groups$V1), max=255), fill=F)}
  polygon3d(x = x_tempo, y = y_tempo, z = z_tempo, col=rgb(vector_color_groups[i,], max=255), fill=F)}

#legend3d("topright", legend = c("Apparition", "Unchanged", "Disapparition"), pch = c(16,16), col = c("red", "black", "blue"), cex=2.5, inset=c(0.02))
arrow3d(c(0,0,0), c(0,5,0), type = "rotation",  col = "green")
arrow3d(c(0,0,0), c(-4,0,0), type = "rotation",  col = "red", cex=4)
arrow3d(c(0,0,0), c(0,0,4), type = "rotation",  col = "blue")
text3d(c(0,6,0), text="N")
text3d(c(-5,0,0), text="W")
text3d(c(0,0,5), text="Z(+)")
axes3d()
title3d(position="topcenter",main = ("Aggregation of polygons"), xlab="x [m]", ylab="y [m]", zlab="z [m]")
htmlwidgets::saveWidget(rglwidget(width = 800, height = 800), 
                        file = "C:/Tile_3D_distribution.html",
                        libdir = "libs",
                        selfcontained = TRUE
)
close3d()
hist(aggregate(data_compiled_assemblages$Surface, by=list(Category=data_groups$V1), FUN=sum)[,2])


path = "C:/"
write.csv(data_groups, paste0(path,"data_groups.csv"), row.names = FALSE)